import numpy as np
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import StandardScaler
from sklearn.datasets import make_moons, make_circles, make_classification
from sklearn.tree import DecisionTreeClassifier
from sklearn.ensemble import RandomForestClassifier
from sklearn.metrics import accuracy_score
import matplotlib.pyplot as plt
import argparse
from matplotlib.colors import ListedColormap



# Function to generate synthetic datasets for classification tasks
def generate_datasets():
    """
    Generates synthetic datasets for classification tasks.

    This function creates three different types of datasets: 
    - Moon-shaped dataset
    - Circle-shaped dataset
    - Linearly separable dataset
    
    Each dataset is suitable for testing the performance of classification algorithms.

    Returns:
        dict: A dictionary containing the datasets. The keys are the names of the datasets
              ('moons', 'circles', 'linearly_separable'), and the values are tuples containing
              the features and labels of the datasets (X, y).
    """
    X, y = make_classification(n_features=2, n_redundant=0, n_informative=2,
                               random_state=1, n_clusters_per_class=1)
    rng = np.random.RandomState(2)
    X += 2 * rng.uniform(size=X.shape)
    linearly_separable = (X, y)

    datasets = {
        "moons": make_moons(noise=0.3, random_state=0),
        "circles": make_circles(noise=0.2, factor=0.5, random_state=1),
        "linearly_separable": linearly_separable
        }
    
    return datasets



def train_classifier_and_evaluate_on_dataset(classifier, datasets: dict):
    """
    Trains a given classifier on multiple datasets and evaluates its performance.

    This function takes a classifier and a dictionary of datasets, trains the classifier on each dataset, 
    and evaluates its performance using accuracy metrics on both the training and testing sets.

    Args:
        classifier (class): A scikit-learn classifier class to be instantiated and trained.
        datasets (dict): A dictionary of datasets where keys are the names of the datasets 
                         ('moons', 'circles', 'linearly_separable'), and values are dictionaries 
                         containing 'X_train', 'X_test', 'y_train', 'y_test'.

    Returns:
        tuple:
            - results (dict): A dictionary where keys are dataset names, and values are dictionaries 
                              containing 'accuracy_train' and 'accuracy_test'.
            - trained_classifiers (dict): A dictionary where keys are dataset names, and values are the 
                                          trained classifier instances.
    """
    # Dictionary to store the evaluation results    
    results = {}
    trained_classifiers = {}

    # Train a Decision Tree Classifier on each dataset and evaluate
    for name, data in datasets.items():
        X_train, X_test, y_train, y_test = data["X_train"], data["X_test"], data["y_train"], data["y_test"]

        # Initialize the classifier and train it on the training data
        cls = classifier().fit(X_train, y_train)    

        # Predict on the training and testing sets
        predictions_test = cls.predict(X_test)
        predictions_train = cls.predict(X_train)    

        # Calculate and store the accuracy for both sets
        accuracy_test = accuracy_score(y_test, predictions_test)
        accuracy_train = accuracy_score(y_train, predictions_train)
        results[name] = {"accuracy_test": accuracy_test, "accuracy_train": accuracy_train}

        # Store the trained classifier
        trained_classifiers[name] = cls
    
    return results, trained_classifiers



def find_optimal_depth_for_classifier(classifier, datasets):
    """
    Finds the optimal depth for a classifier on each dataset and evaluates its performance.

    This function iterates over a range of depth values to find the optimal depth for a given classifier
    on each dataset. It evaluates the classifier's performance based on accuracy on the training and 
    testing sets for each depth.

    Args:
        classifier (class): A scikit-learn classifier class to be instantiated and trained.
        datasets (dict): A dictionary of datasets where keys are dataset names, and values are dictionaries 
                         containing 'X_train', 'X_test', 'y_train', 'y_test'.

    Returns:
        tuple:
            - optimal_depths (dict): A dictionary where keys are dataset names, and values are dictionaries 
                                     containing the optimal depth, and accuracy on the training and testing sets.
            - trained_classifiers (dict): A dictionary where keys are dataset names, and values are the 
                                          trained classifier instances at their optimal depths.
    """
    # Experiment with different 'max_depth' values to find the optimal depth for each dataset
    optimal_depths = {}
    trained_classifiers = {}
    for max_depth in range(1, 10):  # Depth values from 1 to 9
        for name, data in datasets.items():
            X_train, X_test, y_train, y_test = data["X_train"], data["X_test"], data["y_train"], data["y_test"]

            # Train the model with the current max_depth value
            cls = classifier(max_depth=max_depth).fit(X_train, y_train)

            # Evaluate the model on the testing set
            accuracy_test = accuracy_score(y_test, cls.predict(X_test))
            # If the current max_depth gives better accuracy, update the optimal depth and accuracy
            if name not in optimal_depths or optimal_depths[name]["accuracy_test"] < accuracy_test:
                accuracy_train = accuracy_score(y_train, cls.predict(X_train))
                optimal_depths[name] = {"depth": max_depth, "accuracy_test": accuracy_test, "accuracy_train": accuracy_train}
                trained_classifiers[name] = cls
    return optimal_depths, trained_classifiers

def plot_decision_boundaries(X, y, clf, ax, title):
    """
    Plots the decision boundaries of a classifier for a given dataset.

    This function visualizes the decision boundaries of a classifier by plotting the regions of decision 
    along with the data points. It uses different colors to represent different classes and marks the 
    data points with distinct markers.

    Args:
        X (array-like): Feature matrix for the dataset.
        y (array-like): Label vector for the dataset.
        clf (classifier): Trained classifier instance.
        ax (matplotlib.axes.Axes): The axes object to plot the graph on.
        title (str): Title of the plot.
    """
    # Set up marker generator and color map
    markers = ('s', 'x', 'o', '^', 'v')
    colors = ('red', 'blue', 'lightgreen', 'gray', 'cyan')
    cmap = ListedColormap(colors[:len(np.unique(y))])
    
    # Plot the decision surface
    x1_min, x1_max = X[:, 0].min() - 1, X[:, 0].max() + 1
    x2_min, x2_max = X[:, 1].min() - 1, X[:, 1].max() + 1
    xx1, xx2 = np.meshgrid(np.arange(x1_min, x1_max, 0.01),
                           np.arange(x2_min, x2_max, 0.01))
    Z = clf.predict(np.array([xx1.ravel(), xx2.ravel()]).T)
    Z = Z.reshape(xx1.shape)
    ax.contourf(xx1, xx2, Z, alpha=0.4, cmap=cmap)
    ax.set_xlim(xx1.min(), xx1.max())
    ax.set_ylim(xx2.min(), xx2.max())

     # Plot class samples with facecolor to avoid the warning
    # Define filled markers and color map
    markers = ('o', 's', '^', 'v', '<')  # Filled marker styles
    colors = ('red', 'blue', 'lightgreen', 'gray', 'cyan')
    cmap = ListedColormap(colors[:len(np.unique(y))])

    for idx, cl in enumerate(np.unique(y)):
        ax.scatter(x=X[y == cl, 0], 
                   y=X[y == cl, 1],
                   alpha=0.8, 
                   c=colors[idx],
                   marker=markers[idx], 
                   label=cl,
                   edgecolor='black')

    ax.set_title(title)

def plot_input_data(X_train, y_train, X_test, y_test, dataset_name, axs):
    """
    Plot the input data for training and testing datasets on a given axis.

    This function visualizes the training and testing data as scatter plots, 
    using different colors to differentiate between different classes. The 
    function supports 2D data and is intended to be used with subplots.

    Args:
    - X_train (array-like): The training data features, expected to be 2D.
    - y_train (array-like): The training data labels or targets.
    - X_test (array-like): The testing data features, expected to be 2D.
    - y_test (array-like): The testing data labels or targets.
    - dataset_name (str): The name of the dataset, used for titling the plot.
    - axs (matplotlib.axes.Axes): The axes on which the data will be plotted.

    The function uses red and blue colors for the different classes in the dataset,
    with solid colors for the training set and semi-transparent colors for the
    testing set. It sets up the plot limits based on the training data and removes
    the x and y ticks for a cleaner visualization.
    """
    # Set up marker generator and color map
    markers = ('s', 'x', 'o', '^', 'v')
    colors = ('red', 'blue', 'lightgreen', 'gray', 'cyan')
    cmap = ListedColormap(colors[:len(np.unique(y_train))])

    cm_bright = ListedColormap(['#FF0000', '#0000FF'])

    axs.scatter(X_train[:, 0], X_train[:, 1], c=y_train, cmap=cm_bright, edgecolors='k')
    axs.scatter(X_test[:, 0], X_test[:, 1], c=y_test, cmap=cm_bright, alpha=0.6, edgecolors='k')

    axs.set_xlim(X_train[:, 0].min() - 1, X_train[:, 0].max() + 1)
    axs.set_ylim(X_train[:, 1].min() - 1, X_train[:, 1].max() + 1)
    axs.set_xticks(())
    axs.set_yticks(())
    axs.set_title(f"{dataset_name} - Input Data")


def display_results(results, optimal_depths, classifiers_default, classifiers_optimal, datasets):
    """
    Displays the results of classifier evaluations and plots decision boundaries.

    This function visualizes the decision boundaries of classifiers with default and optimal depths
    on both training and testing sets. It creates a grid of plots for each dataset, showing the differences
    in decision boundaries between the default and optimal depth classifiers.

    Args:
        results (dict): Dictionary containing evaluation results of classifiers.
        optimal_depths (dict): Dictionary containing the optimal depths for each dataset.
        classifiers_default (dict): Dictionary containing trained classifiers with default depth.
        classifiers_optimal (dict): Dictionary containing trained classifiers with optimal depth.
        datasets (dict): Dictionary of datasets used for training and testing the classifiers.
    """
    # Visualization setup
    fig, axs = plt.subplots(3, 4+1, figsize=(15, 10))  # Create a grid of subplots
    fig.suptitle('Decision Tree Classifier Boundaries')
    plt.rcParams['font.size'] = 8

    # Create color maps
    cm = plt.cm.RdBu
    cm_bright = ListedColormap(['#FF0000', '#0000FF'])

    # Iterate over the datasets to plot decision boundaries
    for i, (dataset_name, dataset) in enumerate(datasets.items()):
        X_train = dataset["X_train"]
        y_train = dataset["y_train"]
        X_test = dataset["X_test"]
        y_test = dataset["y_test"]

        # Plot input data
        plot_input_data(X_train, y_train, X_test, y_test, dataset_name, axs[i, 0])


        # Add titles and plot decision boundaries for the default depth classifier
        plot_decision_boundaries(X_train, y_train, classifiers_default[dataset_name], axs[i, 1],
                                 title=f"{dataset_name} - Default Depth (Train)")
        plot_decision_boundaries(X_test, y_test, classifiers_default[dataset_name], axs[i, 2],
                                 title=f"{dataset_name} - Default Depth (Test)")

        # Add titles and plot decision boundaries for the optimal depth classifier
        plot_decision_boundaries(X_train, y_train, classifiers_optimal[dataset_name], axs[i, 3],
                                 title=f"{dataset_name} - Optimal Depth (Train)")
        plot_decision_boundaries(X_test, y_test, classifiers_optimal[dataset_name], axs[i, 4],
                                 title=f"{dataset_name} - Optimal Depth (Test)")


        # Remove axis labels and ticks for a cleaner look
        for ax in axs[i]:
            ax.set_xticks([])
            ax.set_yticks([])

    # Adjust the layout and save the figure
    plt.tight_layout(rect=[0, 0.03, 1, 0.95])
    plt.savefig('decision_tree_boundaries.png')
    plt.show()


def main(classifier=DecisionTreeClassifier):
    """
    Main function to run the script: creates the datasets, trains a classifier on each dataset and evaluates it on the training and testing sets.
    and also finds the optimal depth for each dataset and displays the results.
    """
    # Create the datasets
    datasets = generate_datasets()  

    # Preprocess the datasets and split them into training and testing sets
    for name, (X, y) in datasets.items():
        # Standardize features by removing the mean and scaling to unit variance
        X = StandardScaler().fit_transform(X)
        # Split the dataset into a training set and a testing set
        X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=.4, random_state=42)
        datasets[name] = {"X_train": X_train, "X_test": X_test, "y_train": y_train, "y_test": y_test}   

    # Train a Decision Tree Classifier with default parameters on each dataset and evaluate it on the training and testing sets
    results_default_depth, classifiers_default = train_classifier_and_evaluate_on_dataset(classifier, datasets)
    optimal_depths, classifiers_optimal = find_optimal_depth_for_classifier(classifier, datasets)

    display_results(results_default_depth, optimal_depths, classifiers_default, classifiers_optimal, datasets)

if __name__ == "__main__":
    # input classifier name, default is DecisionTreeClassifier
    parser = argparse.ArgumentParser()
    parser.add_argument("--classifier", "-c", default="DecisionTreeClassifier", help="Classifier to use")
    args = parser.parse_args()

    # map classifier name to class
    classifier_dict = {"DecisionTreeClassifier": DecisionTreeClassifier, "RandomForestClassifier": RandomForestClassifier}
    # check if classifier name is valid
    if args.classifier not in classifier_dict:
        print("Invalid classifier name. Valid names are: ", list(classifier_dict.keys()))
        exit(1)

    main(classifier=classifier_dict[args.classifier])
