# Scripts

[[_TOC_]]

## 📁 Overview

This directory contains the scripts used for the project:
- `classify_with_baseline.py` - Classify images using a random and a majority classifier.
- `classify_with_basic_classifiers.py` - Classify images using a decision tree, random forest or a naive bayes classifier.
  - `find_optimal_parameters_basic.py` - Helper script to find the optimal parameters for a decision tree, random forest or a naive bayes classifier.
- `classify_with_CNN.py` - Classify images using a convolutional neural network.
  - `evaluate_cnn.py` Evaluate the trained models and create plots


## ✔ Setup

- tested with Python 3.11 (other versions might work as well)

Create a virtual environment and install the required packages:

```bash
python3.11 -m venv venv
source venv/bin/activate
pip install -r requirements.txt
```

## `classify_with_baseline.py`

The script `classify_with_baseline.py` classifies our image dataset using a random and a majority classifier. The random classifier assigns a random class to each image, while the majority classifier assigns the majority class (= *apple*) to each image.

### 💻 Usage

```bash
python classify_with_baseline.py [--include_sklearn]
```

| Parameter | Description | 
| --------- | ----------- | 
| `--include_sklearn` | Whether to include baseline implementation from scikit-learn in the plot. | 

### 📊 Outputs

Each run of the script will ...
- write the accuracy, micro precision, micro recall and micro F1 scores, macro precision, macro recall and macro F1 scores to the console
- plot the evaluation results for the training, dev and test set in a bar chart and save it to `../figures/baselines.png`
- plot the class distribution of the training set as a histogram and save it to `../figures/class_distribution_histogram_train.png` (because the class distribution decides the majority class)


## `classify_with_basic_classifiers.py`

The script `classify_with_basic_classifiers.py` is used to classify images using a decision tree, random forest or a naive bayes classifier. The script can be used to classify images using different filters and resizing the images to different image sizes. Additionally, the optimal parameters for the classifiers can be found using GridSearch. The script also allows for training on different subsets of the training data.

### 💻 Usage

```bash
python classify_with_basic_classifiers.py --classifier <classifier_name> [--filters <filters>] [--resize <resize-1> <resize-2>] [--fraction <fraction>] [--optimize] [--debug]
```

**Example call:**
- Classify images using a decision tree classifier with the HSV and sobel filters and find the optimal parameters.

```bash
python classify_with_basic_classifiers.py -c DecisionTree --filters hsv sobel -o
```

| Parameter | Description | Choices / Examples |
| --------- | ----------- | ------- |
| `--classifier` / `-c` | The classifier to use. | `DecisionTree`, `RandomForest`, `NaiveBayes` |
| `--filters` / `-f` | The filters to use. To use several filters, separate them by a space. Default: `standard`.  | `standard` => no filters, `sobel`, `gray`, `hsv`, `canny` |
| `--resize` / `-r ` | The size to resize the images to. Default: `50 50`. | e.g. `100 100` |
| `--fraction` / `-fr` | The fraction of the dataset to train the classifier on. Default: `100`. | e.g. `50`  to use a random half of the train split | 
| `--optimize` / `-o` | Whether to find the optimal parameters for the classifier with GridSearch or use predefined parameters in the script. | |
| `--debug` / `-d` | Whether to enable debug logging. | |


### 📊 Outputs

Each run of the script will ...
- store the features and labels in a pickle file in the  `../features` directory, if the file does not exist yet.
- write the accuracy, precision, recall and F1 scores to the console and to a simple text file `results_basic_classifiers.txt`
- save the **confusion matrix** to the respective classifier's directory in `../figures/`.
- save a heatmap showing the **feature importances** to the respective classifier's directory in `../figures/` (for random forest and decision tree classifiers).
- if the `--optimize` flag is set, the optimal hyperparameters will be found and a plot with the GridSearch results will be generated and saved.
- save the trained classifier to a pickle file in the `../trained_classifiers` directory.

## `classify_with_cnn.py`

The script `classify_with_cnn.py` is designed to train a Convolutional Neural Network (CNN) for image classification. It includes functions for image loading, CNN model creation, training, and evaluation. The script allows for training on different subsets of the training data.

### 💻 Usage

```bash
python classify_with_vnn.py [--train_subsets <proportions>]
```

| Parameter | Description | Choices / Examples |
| --------- | ----------- | ------------------ |
| `--train_subsets` | List of proportions of the training data to use. | `[0.1, 0.5, 1.0]` (example) |

### 📊 Outputs

For each specified training subset, the script will:
- Train a CNN model on the subset of the training data.
- Plot and save the training and validation loss over epochs to `../figures/cnn/fruit_classifier_<subset_description>/loss_plot_<subset_description>.png`.
- Plot and save the training and validation accuracy over epochs to `../figures/cnn/fruit_classifier_<subset_description>/accuracy_plot_<subset_description>.png`.
- Save the trained CNN model to `../trained_classifiers/fruit_classifier_<subset_description>.keras`.
- Save the label encoder to `../figures/cnn/label_encoder.pkl`. 


## `evaluate_cnn.py`

The `evaluate_cnn.py` script allows for evaluating CNN models on specific subsets of the training data via a command-line parameter. This enhancement increases its flexibility for different model evaluations.

### 💻 Usage

```bash
python evaluate_cnn.py <subset>
```

| Parameter | Description | Choices / Examples |
| --------- | ----------- | ------------------ |
| `subset` | Specifies the subset of the training data used for the model. | `'10_percent'`, `'50_percent'`, `'full'` |

### 📊 Outputs

For the specified model subset, the script will:
- Print accuracy, precision (macro), recall (macro), and F1 score (macro) to the console for each data split (train, dev, test).
- Save plots of these metrics in `../figures/cnn/fruit_classifier_<subset>/`.
- Save a confusion matrix for each data split in the same directory.
- Display and save images of misclassified samples from the test set.
