"""
Random Baseline & Majority Baseline
"""

import argparse
import os
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
import random
from sklearn.dummy import DummyClassifier
from sklearn.metrics import accuracy_score
from sklearn.metrics import precision_recall_fscore_support

classes = [
    "apple", "apricot", "banana", "blueberry", "buddha's hand", "cherry", 
    "coconut", "date", "dragonfruit", "durian", "fig", "grape", 
    "grapefruit", "jostaberry", "kiwi", "lemon", "lime", "lychee", 
    "mandarine", "mango", "nectarine", "orange", "passion fruit", "pear", 
    "pineapple", "pomegranate", "raspberry", "redcurrant", "strawberry", "watermelon"
]

train_dir = "../data/train"

def get_class_distribution(dir: str):
    """
    Get the distribution of classes in a directory
    """
    class_distribution = {}
    count = 0
    for cls in classes:
        class_distribution[cls] = len(os.listdir(os.path.join(dir, cls)))
        count += class_distribution[cls]
    print(f"Total number of images in training set: {count}")
    return class_distribution, count

def create_histogram(distribution: dict, save_path: str):
    """
    Create a histogram of the distribution of classes
    """
    df = pd.DataFrame(list(distribution.items()), columns=['Class', 'Number of Images'])
    plt.figure(figsize=(12, 8))
    plt.bar(df['Class'], df['Number of Images'], color='skyblue')
    plt.xlabel('Class')
    plt.ylabel('Number of Images')
    plt.title('Histogram of Image Distribution Across Classes in Entire Dataset')
    plt.xticks(rotation=90)
    plt.subplots_adjust(bottom=0.3)  # Adjusting bottom to fit long names
    plt.savefig(save_path)
    print(f"Histogram saved to {save_path}")


def load_data(split):
    """
    Load data from a split and prepare for DummyClassifier.

    Args:
        split: train, dev or test.

    Returns:
        X: List of features (dummy in this case).
        y: List of labels.
    """
    data_dir = f"../data/{split}"
    X = []
    y = []

    for cls in classes:
        num_instances = len(os.listdir(os.path.join(data_dir, cls)))
        X.extend([[0]] * num_instances)  # Dummy features, as they are not used
        y.extend([cls] * num_instances)
    
    return X, y

def run_dummy_classifier(strategy, train_X, train_y, data_X):
    """
    Run the DummyClassifier with a given strategy and return predictions.

    Args:
        strategy: The strategy for the DummyClassifier ('uniform' or 'most_frequent').
        train_X, train_y: Training features and labels.
        data_X: Testing features.

    Returns:
        predictions: Predictions made by the classifier.
    """
    dummy_clf = DummyClassifier(strategy=strategy)
    dummy_clf.fit(train_X, train_y)
    predictions = dummy_clf.predict(data_X)
    return predictions



def random_baseline(probabilities: dict, split: str):
    """
    Randomly predict a class based on the distribution of classes in the train set
    
    Args:
        probabilities: Dictionary of probabilities of each class
        split: train, dev or test
        
    Returns:
        predictions: Dictionary of predictions for each class
        """
    data_dir = f"../data/{split}"
    predictions = []

    for cls in classes:
        num_instances = len(os.listdir(os.path.join(data_dir, cls)))
        predictions.extend(random.choices(classes, weights=list(probabilities.values()), k=num_instances))
    return predictions

    
def majority_baseline(train_distribution: dict, split: str):
    """
    Assign the most common class in the training set to all instances in the given split.

    Args:
        train_distribution: Distribution of classes in the training set.
        split: train, dev or test.

    Returns:
        predictions: Dictionary of predictions for each class.
    """
    # Find the most common class
    most_common_class = max(train_distribution, key=train_distribution.get)
    
    # Assign this class to all instances in the split
    data_dir = f"../data/{split}"
    predictions = []
    for cls in classes:
        num_instances = len(os.listdir(os.path.join(data_dir, cls)))
        predictions.extend([most_common_class] * num_instances)
    return predictions

def evaluate_baselines(true_labels, predictions_dict, split):
    """
    Evaluate baselines using precision, recall, and F1-score.

    Args:
        true_labels: The true labels of the test set.
        predictions_dict: A dictionary containing the predictions from each baseline.
                          Each key is the name of the baseline, and the value is the predictions.
        split: train, dev or test

    Returns:
        evaluation_dict: A dictionary containing the evaluation results for each baseline.
                         Each key is the name of the baseline, and the value is a dictionary
                         containing the precision, recall, and F1-score for both micro and macro averaging.
    """
    evaluation_dict = {}

    for baseline_name, predictions in predictions_dict.items():
        # Calculate precision, recall, and F1-score for both micro and macro averaging
        accuracy = accuracy_score(true_labels, predictions)
        precision_macro, recall_macro, f1_macro, _ = precision_recall_fscore_support(true_labels, predictions, average='macro', zero_division=0)
        precision_micro, recall_micro, f1_micro, _ = precision_recall_fscore_support(true_labels, predictions, average='micro', zero_division=0)

        # Print the results
        print(f"Results for {baseline_name} -- {split.upper()}:")
        print(f"  Accuracy: {accuracy:.3f}")
        print(f"  Macro Average - Precision: {precision_macro:.3f}, Recall: {recall_macro:.3f}, F1-Score: {f1_macro:.3f}")
        print(f"  Micro Average - Precision: {precision_micro:.3f}, Recall: {recall_micro:.3f}, F1-Score: {f1_micro:.3f}\n\n")

        # Add the results to the dictionary
        evaluation_dict[baseline_name] = {
            "accuracy": accuracy,
            "precision_macro": precision_macro,
            "precision_micro": precision_micro,
            "recall_macro": recall_macro,
            "recall_micro": recall_micro,
            "f1_macro": f1_macro,
            "f1_micro": f1_micro
        }
    return evaluation_dict
    

def plot_baselines(evaluation: dict):
    """
    Plot the results of the baselines in a bar chart and save to figures/baselines.png.

    Args:
        evaluation: A dictionary containing the evaluation results for each baseline.
                    Each key is the name of the baseline, and the value is a dictionary
                    containing the precision, recall, and F1-score for both micro and macro averaging.
    
    """
    fig, axes = plt.subplots(nrows=1, ncols=3, figsize=(26, 13))

    # Plot the results for each split in a separate subplot
    for i, split in enumerate(['train', 'dev', 'test']):
        # Create pandas dataframe
        df = pd.DataFrame.from_dict(evaluation[split], orient='index')
        df.plot(kind='bar', y=["accuracy", "precision_macro", "precision_micro", "recall_macro", "recall_micro", "f1_macro", "f1_micro"], rot=0, ax=axes[i],
                title=split, ylabel="Score", legend=True, colormap='Spectral', fontsize=14)

    fig.suptitle('Evaluation of Baselines', fontsize=22)
    fig.tight_layout(pad=2.8)
    
    plt.savefig("../figures/baselines.png")

    print("Plot saved to ../figures/baselines.png")
    plt.show()


def main(include_sklearn):
    """
    Run the baselines and plot the results.

    Args:
        include_sklearn: Whether to include the sklearn baselines in the plot.
    """
    # Get distribution of classes in train set
    distribution_train, num_train = get_class_distribution(train_dir)  

    # Create histogram of distribution
    create_histogram(distribution_train, "../figures/class_distribution_histogram_train.png")  

    # Divide by total number of images to get probabilities
    probabilities_train = {k: v/num_train for k, v in distribution_train.items()}   

    train_X, train_y = load_data('train')

    predictions_dict = {}
    evaluation_dict = {}

    for split in ['train', 'dev', 'test']:
        random_predictions = random_baseline(probabilities_train, split)
        majority_predictions = majority_baseline(distribution_train, split)
        data_X, data_y = load_data(split) 
        
        random_predictions_sklearn = run_dummy_classifier('stratified', train_X, train_y, data_X)
        majority_predictions_sklearn = run_dummy_classifier('most_frequent', train_X, train_y, data_X)

        # Store predictions in dictionary
        predictions_dict[split] = {
        "Random/Stratified (Custom)": random_predictions,
        "Random/Stratified (Sklearn)": random_predictions_sklearn,
        "Majority (Custom)": majority_predictions,
        "Majority (Sklearn)": majority_predictions_sklearn
        }

        evaluation_dict[split] = evaluate_baselines(data_y, predictions_dict[split], split)


    if include_sklearn:
        plot_baselines(evaluation_dict)
    else:
        for split in ['train', 'dev', 'test']:
            evaluation_dict[split].pop("Random/Stratified (Sklearn)")
            evaluation_dict[split].pop("Majority (Sklearn)")
        
        plot_baselines(evaluation_dict)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Run baselines")
    parser.add_argument("--include_sklearn", action="store_true", help="Include sklearn baselines in plot")
    args = parser.parse_args()

    main(args.include_sklearn)