"""Classify images using 3 different classifiers (Decision Tree, Random Forest, Naive Bayes)."""

# General imports
import argparse
import cv2 # requirement: pip install opencv-python
import logging
import matplotlib.pyplot as plt
import numpy as np
import os
import pickle
import random
import time

# Image feature extraction
from skimage.filters import sobel # requirement: pip install scikit-image

# Classifiers from scikit-learn 
from sklearn.tree import DecisionTreeClassifier
from sklearn.ensemble import RandomForestClassifier
from sklearn.naive_bayes import GaussianNB

from sklearn.tree import plot_tree

# Metrics
from sklearn.metrics import accuracy_score, confusion_matrix, precision_recall_fscore_support, ConfusionMatrixDisplay


# Local imports for hyperparameter optimization
from find_optimal_parameters_basic import find_best_params_for_decision_tree, find_best_params_for_random_forest, find_best_params_for_naive_bayes


def read_and_resize_image(image_path: str, resize: tuple) -> np.ndarray:
    """
    Reads an image from a file and resizes it.

    Args:
        image_path (str): Path to the image file.
        resize (tuple): The desired image size after resizing.


    Returns:
        image (np.ndarray): The resized image.
    """
    image = cv2.imread(image_path)
    if image is None:
        raise ValueError(f"Could not read image: {image_path}")
    
    logging.debug(f"Shape of image before resizing: {image.shape}")
    image = cv2.resize(image, resize) # e.g. resize = (50, 50) for 50x50 pixels
    logging.debug(f"Shape of image after resizing: {image.shape}")
    
    # Save resized image (not yet sorted by class); optionally uncomment the following lines to save the resized images
    """
    image_name = image_path.split("/")[-1]
    #cv2.imwrite("../resized_images/" + image_name, image)
    """
    
    return image


def apply_filters(image: np.ndarray, filters: list) -> dict:
    """
    Applies a list of filters to an image and returns the extracted features.

    Args:
        image (np.ndarray): The image from which to extract features.
        filters (list): The types of features to extract.

    Returns:
        features (dict): A dictionary containing the extracted features.
    """    
    features = {}
    if "standard" in filters:
        features["standard"] = image.flatten()
    if "gray" in filters:
        image_gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY) # Grayscale image
        features["gray"] = image_gray.flatten()
    if "sobel" in filters:
        image_sobel = sobel(image) # Edge detection but with color information
        features["sobel"] = image_sobel.flatten()
    if "hsv" in filters:
        image_hsv = cv2.cvtColor(image, cv2.COLOR_BGR2HSV) # HSV (hue, saturation, value) color space
        features["hsv"] = image_hsv.flatten()
    if "canny" in filters:
        image_canny = cv2.Canny(image, threshold1=150, threshold2=300) # Canny edge detection; thredholds are roughly fine-tuned by hand
        features["canny"] = image_canny.flatten()
    
    logging.debug(f"No. of features: {len(features)}")
    return features


def extract_features_from_image(image_path: str, resize: tuple, filters: list) -> np.ndarray:
    """
    Extracts features from an image file by first resizing the image and then applying a list of filters.

    Args:
        image_path (str): Path to the image file.
        resize (tuple): The desired image size after resizing.
        filters (list): The types of features to extract.

    Returns:
        features (np.ndarray): A flat vector containing the extracted features.
    """

    image = read_and_resize_image(image_path, resize)
    features = apply_filters(image, filters)
    features = np.concatenate([features[key] for key in features.keys()])
    return features
    

def load_and_prepare_dataset(dataset_path: str, backup_path: str, resize: tuple, name: str, filters: list, fraction: float = 100.0) -> dict:
    """
    Loads the datasets for our classification task, extracts the features from the images and stores the dataset in a pickle file for faster access in next runs.
    
    If the pickle file already exists, the dataset is loaded from the pickle file.

    Args:
        dataset_path (str): Path to the dataset. Needs to contain a 'train', 'dev' and 'test' folder.
        backup_path (str): Path to the pickle file where the dataset will be stored.
        resize (tuple): Tuple containing the desired image size.
        name (str): Name that characterizes the extracted features (e.g. "sobel")
        filters (list): List of filters that should be applied to the images (e.g. ["sobel"] or ["gray", "sobel"])

    Returns:
        dataset (dict): A dictionary containing the datasets for each split. Keys are: 'train', 'dev' and 'test'.
    """
    
    # If the pickle file already exists, load the dataset from it
    if os.path.isfile(backup_path):
        with open(backup_path, "rb") as f:
            logging.info(f"Loading dataset from pickle file {backup_path}")
            dataset = pickle.load(f)
            # For debugging, show the shape of the first image
            logging.debug(f"Shape of first image: {dataset['train'][0][0].shape}")
            return dataset

    # Create directory for storing features if it does not exist yet
    if not os.path.exists("../features"):
        os.makedirs("../features")
        
    dataset = {"train": [], "dev": [], "test": []}

    # Iterate over dataset splits
    for split in dataset.keys():
        # Iterate over classes
        for class_name in os.listdir(f"{dataset_path}/{split}"):
            logging.info(f"Loading {split} dataset, class: {class_name}")
            # Get list of all images for this class
            all_images = os.listdir(f"{dataset_path}/{split}/{class_name}")
            if split == "train" and fraction < 100.0:
                # Select a random fraction of the images
                selected_images = random.sample(all_images, int(len(all_images) * fraction / 100))
            else:
                selected_images = all_images
            # Iterate over images
            for image_id in selected_images:
                # Extract features from image
                features = extract_features_from_image(f"{dataset_path}/{split}/{class_name}/{image_id}", resize, filters)
                # Add features and label to dataset
                X_y = (features, class_name)
                dataset[split].append(X_y)
        print(f"Number of images in {split} set: {len(dataset[split])}")
          
    # Store dataset in pickle file
    with open(backup_path, "wb") as f:
        pickle.dump(dataset, f)
        logging.info(f"Saved dataset to pickle file {backup_path}")
    
    return dataset

def plot_feature_importances(cls, feature_description: str, resize: tuple, params_as_string: str, cls_name: str):
    """
    Plots the feature importances for a Decision Tree or Random Forest classifier.
    
    Args:
        cls (class): The scikit-learn classifier class that was trained.
        feature_description (str): A string describing the features used for training the classifier.
        resize (tuple): The image size after resizing.
        params_as_string (str): A string containing the parameters used for training the classifier.
        cls_name (str): The name of the classifier.
    """

    # Get feature importances from the classifier
    feature_importances = cls.feature_importances_
    image_feature_importances = feature_importances.reshape(resize[0], resize[1], 3)
    
    # Calculate the mean feature importance across color channels because we have 3 color channels but we only want one value per pixel
    feature_importances_per_pixel = np.mean(image_feature_importances, axis=2)
    
    mean_feature_importances = np.mean(feature_importances_per_pixel)
    logging.debug(f"Average feature importance: {mean_feature_importances}")
  
    # Plot the feature importances with a heatmap
    plt.figure(figsize=(10,10))
    
    
    plt.imshow(feature_importances_per_pixel, cmap=plt.cm.hot)
    plt.title(f"Feature importances for {cls_name} classifier on {feature_description} images")
    plt.xlabel("Pixel")
    plt.ylabel("Pixel")
    plt.colorbar()
    plt.savefig(f"../figures/{cls_name}/{cls_name}_{feature_description}_feature_importances_{params_as_string}.png")
    logging.info(f"Feature importances saved to ../figures/{cls_name}/{cls_name}_{feature_description}_feature_importances_{params_as_string}.png")


def train_and_evaluate_classifier(classifier, dataset, feature_description, resize=(50, 50), optimize=False):
    """
    Trains a classifier (Decision Tree, Random Forest or Naive Bayes) on the dataset and if optimize=True,
    finds the optimal depth for the classifier, and evaluates its performance.

    The classifier's performance is evaluated based on accuracy on the dev set.

    Args:
        classifier (class): A scikit-learn classifier class to be instantiated and trained.
        dataset (dict): A dictionary where keys are 'train', 'dev' and 'test' and values are tuples with features and respective labels (= class names).
        feature_description (str): A string describing the features used for training the classifier.
        resize (tuple): The image size after resizing.
        optimize (bool): Whether to find the optimal parameters for the classifier.
    """

    # Assign the datasets to variables
    X_train, X_dev, X_test, y_train, y_dev, y_test = list(map(lambda x: x[0], dataset["train"])), list(map(lambda x: x[0], dataset["dev"])), list(map(lambda x: x[0], dataset["test"])), list(map(lambda x: x[1], dataset["train"])), list(map(lambda x: x[1], dataset["dev"])), list(map(lambda x: x[1], dataset["test"]))


    # Either optimize the classifier or use predefined parameters
    if optimize:
        logging.info(f"Finding the optimal depth for {classifier.__name__} classifier on the dataset...")
        start_time = time.time()

        # Find the optimal parameters for the classifier using GridSearchCV in a separate module
        if classifier.__name__ == "DecisionTreeClassifier":
            params = find_best_params_for_decision_tree(X_train, y_train, X_dev, y_dev, feature_description)
        elif classifier.__name__ == "RandomForestClassifier":
            params = find_best_params_for_random_forest(X_train, y_train, X_dev, y_dev, feature_description)
        elif classifier.__name__ == "GaussianNB":
            params = find_best_params_for_naive_bayes(X_train, y_train, X_dev, y_dev, feature_description)
        else:
            print(classifier.__name__ + " is not supported yet")
        logging.info(f"Optimal parameters for {classifier.__name__}: {params}")
        logging.info(f"Time for finding optimal parameters: {((time.time() - start_time) / 60):.5f} minutes")
        
    else:
        # If we do not optimize, we want use the following parameters for the classifier because they are on average the best parameters
        # Change them if you want to experiment with different parameters
        if classifier.__name__ == "DecisionTreeClassifier":
            params = {"max_depth": 60, "max_features": "sqrt", "min_samples_leaf": 1, "min_samples_split": 2} # best for HSV + Sobel
        elif classifier.__name__ == "RandomForestClassifier":
            params = {"max_depth": 40, "max_features": "sqrt", "min_samples_leaf": 2, "min_samples_split": 2, "n_estimators": 150} # best for HSV
        elif classifier.__name__ == "GaussianNB":
            params = {"var_smoothing": 4.2e-08} # best for HSV + Sobel
        
        logging.info(f"Using the following parameters for {classifier.__name__}: {params}")
    
    
    params_as_string = "_".join([str(key) + "_" + str(params[key]) for key in params.keys()]) # needed for saving the results and plots

    # Instantiate the classifier with the optimal/predefined parameters
    start_time = time.time()
    if classifier.__name__ == "GaussianNB":
        cls = classifier(**{key: params[key] for key in params})
    else:
        cls = classifier(random_state=42, **{key: params[key] for key in params})
    
    # Fit the classifier on the training set; we need to do this here because GridSearchCV would refit it on the training + dev set
    cls.fit(X_train, y_train)
    logging.info(f"Time for training {classifier.__name__}: {((time.time() - start_time) / 60):.5f} minutes")
    
    # Save the trained classifier
    if not os.path.exists("../trained_classifiers"):
        os.makedirs("../trained_classifiers")
    logging.info(f"Saving trained classifier to trained_classifiers/{classifier.__name__}_{feature_description}_{params_as_string}.pkl")
    pickle.dump(cls, open(f"../trained_classifiers/{classifier.__name__}_{feature_description}_{params_as_string}.pkl", "wb"))
    
    
    # Plotting the decision tree is not very useful because the tree is quite large, but if you want to do it, uncomment the following lines
    """
    if classifier.__name__ == "DecisionTreeClassifier":
        feature_names = ["pixel_" + str(i) for i in range(resize[0] * resize[1] * 3)]
        fig, ax = plt.subplots(figsize=(22, 10))
        plot_tree(cls, filled=True, impurity=False, feature_names=feature_names, class_names=cls.classes_, rounded=True, proportion=False, precision=1, fontsize=10, ax=ax)
        plt.title(f"Decision Tree for {feature_description} images")
        plt.savefig(f"../figures/decision_tree/plotted_tree_{feature_description}_{params_as_string}.png")
    """
    
    # Dictionary to map classifier names to shorter names for saving plots
    classifier_name = {"DecisionTreeClassifier": "decision_tree", "RandomForestClassifier": "random_forest", "GaussianNB": "naive_bayes"}
    
    # Plot feature importances for decision tree and random forest
    if classifier.__name__ == "RandomForestClassifier" or classifier.__name__ == "DecisionTreeClassifier":
        plot_feature_importances(cls, feature_description, resize, params_as_string, classifier_name[classifier.__name__])
    
    # Predict on the training, dev and testing sets and calculate the accuracy scores
    logging.info(f"Predicting on the training, dev and testing sets and calculating the accuracy scores ...")
    predictions_train = cls.predict(X_train)
    predictions_dev = cls.predict(X_dev)
    predictions_test = cls.predict(X_test)

    accuracy_train = accuracy_score(y_train, predictions_train)
    accuracy_dev = accuracy_score(y_dev, predictions_dev)
    accuracy_test = accuracy_score(y_test, predictions_test)

    precision_macro, recall_macro, f1_macro, _ = precision_recall_fscore_support(y_dev, predictions_dev, average="macro")
    precision_micro, recall_micro, f1_micro, _ = precision_recall_fscore_support(y_dev, predictions_dev, average="micro")
    
    
    results = {"accuracy_train": accuracy_train, "accuracy_dev": accuracy_dev, "accuracy_test": accuracy_test,
                "precision_macro": precision_macro, "recall_macro": recall_macro, "f1_macro": f1_macro,
                "precision_micro": precision_micro, "recall_micro": recall_micro, "f1_micro": f1_micro}
    
    logging.info(f"Accuracy on training set: {accuracy_train}")
    logging.info(f"Accuracy on dev set: {accuracy_dev}")
    logging.info(f"Accuracy on testing set: {accuracy_test}")
    logging.info(f"Precision macro dev: {precision_macro}, Recall macro dev: {recall_macro}, F1 macro dev: {f1_macro}")
    logging.info(f"Precision micro dev: {precision_micro}, Recall micro dev: {recall_micro}, F1 micro dev: {f1_micro}")

    # Plot the confusion matrix
    fig, ax = plt.subplots(figsize=(20, 20))
    #cm = confusion_matrix(y_dev, predictions_dev, labels=cls.classes_)
    
    ConfusionMatrixDisplay.from_estimator(estimator=cls, X=X_dev, y=y_dev, display_labels=cls.classes_, xticks_rotation="vertical", ax=ax)
    
    plt.title(f"Confusion matrix for {classifier.__name__} classifier on {feature_description} images")
    plt.savefig(f"../figures/{classifier_name[classifier.__name__]}/{classifier.__name__}_{feature_description}_confusion_matrix_{params_as_string}.png")
    logging.info(f"Confusion matrix saved to ../figures/{classifier_name[classifier.__name__]}/{classifier.__name__}_{feature_description}_confusion_matrix_{params_as_string}.png")
        
    # Save the results to a simple text file
    with open ("results_basic_classifiers.txt", "a") as f:
        f.write(f"Results for {classifier.__name__} classifier on {feature_description} images:\n")
        f.write(f"Optimal parameters: {params}\n")
        f.write(str(results) + "\n\n")
    
    
    # Store some images with their predicted and true labels
    # Create directories if they don't exist yet
    if not os.path.exists("../figures/correctly_classified"):
        os.makedirs("../figures/correctly_classified")
    if not os.path.exists("../figures/misclassified"):
        os.makedirs("../figures/misclassified")
    
    for i in range(11):
        image_id = random.randint(0, len(X_dev))
        if predictions_dev[image_id] == y_dev[image_id]:
            dir_name = "correctly_classified"
        else:
            dir_name = "misclassified"
        size = resize[0] * resize[1] * 3
        # DISCLAIMER: this does not work with every combination of filters because it depends on the order of the filters and the size of the feature vector
        if "hsv" in feature_description:
            image = X_dev[image_id][:size].reshape(resize[0], resize[1], 3)
            cv2.imwrite(f"../figures/{dir_name}/{classifier.__name__}_{feature_description}_predicted_{predictions_dev[image_id]}_true_{y_dev[image_id]}_{params_as_string}.png", image)
        if "standard" in feature_description:
            image = X_dev[image_id][:size].reshape(resize[0], resize[1], 3)
            cv2.imwrite(f"../figures/{dir_name}/{classifier.__name__}_{feature_description}_predicted_{predictions_dev[image_id]}_true_{y_dev[image_id]}_{params_as_string}.png", image)
        if "sobel" in feature_description:
            image_sobel = X_dev[image_id][size:2*size].reshape(resize[0], resize[1], 3)
            cv2.imwrite(f"../figures/{dir_name}/{classifier.__name__}_{feature_description}_predicted_{predictions_dev[image_id]}_true_{y_dev[image_id]}_{params_as_string}_sobel.png", image_sobel * 255)
        
def main(classifier=DecisionTreeClassifier, resize=(50, 50), filters=["standard"], optimize=False, fraction=100.0):
    """
    Main function to run the classifier: preprocesses the dataset splits, trains a classifier on the dataset and evaluates it on the dev and test set.
    
    It optionally also finds the optimal parameters using GridSearch.
    
    Args:
        classifier (class): A scikit-learn classifier class to be trained: DecisionTreeClassifier, RandomForestClassifier or GaussianNB.
        resize (tuple): The desired image size after resizing.
        filters (list): The types of features to extract.
        optimize (bool): Whether to find the optimal parameters for the classifier.
        fraction (float): Fraction of the train dataset to use for training.
    """

    # ------- Prepare the dataset ---------
    name = "_".join(sorted(filters)) # e.g. "hsv_sobel"
    
    # Set the pickle filename for the dataset where the features are stored
    if fraction == 100.0:
        pickle_filename = f"../features/dataset_resized_{resize[0]}x{resize[1]}_{name}.pickle"
    else:
        pickle_filename = f"../features/dataset_resized_{resize[0]}x{resize[1]}_{name}_{str(fraction)}.pickle"
        name += f"_{str(fraction)}"

    # Load dataset and extract features
    dataset = load_and_prepare_dataset("../data", pickle_filename, resize, name, filters, fraction)

    # Check if the dataset has the expected number of feature vectors
    if fraction == 100.0:
        num_feature_vectors = len(dataset["train"]) + len(dataset["dev"]) + len(dataset["test"])
        assert num_feature_vectors == 29430, f"Expected 29430 images, but got {num_feature_vectors}"
    
    # ------- Train and evaluate the classifier ---------
    train_and_evaluate_classifier(classifier, dataset, f"{resize[0]}x{resize[1]}_{name}", resize, optimize)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter) # Show default values
    parser.add_argument("--classifier", "-c", choices=["DecisionTree", "RandomForest", "NaiveBayes"], help="Classifier/model to use")
    parser.add_argument("--filters", "-f", nargs="+", default=["standard"], choices=["standard", "sobel", "gray", "hsv", "canny"], help="Filters to use")
    parser.add_argument("--resize", "-r", nargs=2, default=(50, 50), type=int, help="Resize images to this size")
    parser.add_argument("--optimize", "-o", action="store_true", help="Find optimal parameters for classifier")
    parser.add_argument("--fraction", "-fr", default=100.0, type=float, help="Fraction of the train dataset to use for training")
    parser.add_argument("--debug", "-d", action="store_true", help="Enable debug logging")
    
    args = parser.parse_args()

    # Map classifier name to class
    classifier_dict = {
        "DecisionTree": DecisionTreeClassifier,
        "RandomForest": RandomForestClassifier,
        "NaiveBayes": GaussianNB
        }

    # Set logging level
    logging.basicConfig(level=logging.DEBUG) if args.debug else logging.basicConfig(level=logging.INFO)

    # Run the main function
    main(classifier_dict[args.classifier], args.resize, args.filters, args.optimize, args.fraction)
