import os
import cv2
import numpy as np
import argparse
from sklearn.preprocessing import LabelEncoder
from tensorflow.keras import layers, models, optimizers
from tensorflow.keras.utils import to_categorical
from tensorflow.keras.preprocessing.image import ImageDataGenerator
from tensorflow.keras.callbacks import LearningRateScheduler
import matplotlib.pyplot as plt
import seaborn as sns
from sklearn.model_selection import train_test_split
import joblib

# Function to load images and labels for CNN
def load_images_and_labels_cnn(base_path, size=(50, 50)):
    """
    Load images and their labels for CNN processing from a given base path.
    Images are resized to a specified size.

    Parameters:
    base_path (str): The path of the directory containing image folders.
    size (tuple): The desired size for resizing the images (width, height).

    Returns:
    tuple: A tuple containing arrays of images and labels.
    """

    images = []
    labels = []
    for folder in os.listdir(base_path):
        folder_path = os.path.join(base_path, folder)
        for image_file in os.listdir(folder_path):
            img_path = os.path.join(folder_path, image_file)
            img = cv2.imread(img_path)
            if img is not None:
                img = cv2.resize(img, size)
                images.append(img)
                labels.append(folder)
            else:
                print("Error loading image: " + img_path)
    return np.array(images), np.array(labels)

# Function to create the CNN model
def get_model(lr=0.001, dropout_rate=0.3, activation='relu'):
    """
    Create and compile a Convolutional Neural Network (CNN) model.

    Parameters:
    lr (float): Learning rate for the optimizer. Default is 0.001.
    dropout_rate (float): Dropout rate for regularization. Default is 0.3.
    activation (str): Activation function for the layers. Default is 'relu'.

    Returns:
    tensorflow.keras.models.Sequential: The compiled CNN model.
    """

    input_shape = (50, 50, 3)
    outputs_number = 30
    model = models.Sequential([
        layers.InputLayer(input_shape=input_shape),
        layers.Conv2D(64, (3, 3), activation=activation, padding='same'),
        layers.MaxPooling2D((2, 2)),
        layers.Conv2D(128, (3, 3), activation=activation, padding='same'),
        layers.MaxPooling2D((2, 2)),
        layers.Conv2D(256, (3, 3), activation=activation, padding='same'),
        layers.MaxPooling2D((2, 2)),
        layers.GlobalAveragePooling2D(),
        layers.Dense(256, activation=activation),
        layers.Dropout(dropout_rate),
        layers.Dense(outputs_number, activation='softmax')
    ])
    optimizer = optimizers.Adam(learning_rate=lr)
    model.compile(loss='categorical_crossentropy', optimizer=optimizer, metrics=['accuracy'])
    return model

# Learning rate schedule function
def lr_schedule(epoch):
    """
    Learning rate schedule function to adjust the learning rate based on the epoch.

    Parameters:
    epoch (int): The current epoch number during training.

    Returns:
    float: The adjusted learning rate.
    """

    lr = 0.001
    if epoch > 50:
        lr *= 0.5e-3
    elif epoch > 40:
        lr *= 1e-3
    elif epoch > 30:
        lr *= 1e-2
    elif epoch > 20:
        lr *= 1e-1
    return lr

# Function to create stratified subsets of data
def stratified_subset(X, y, subset_ratio):
    """
    Create a stratified subset of data based on the given subset ratio.

    Parameters:
    X (numpy.ndarray): The input features.
    y (numpy.ndarray): The target labels.
    subset_ratio (float): The ratio of the subset size to the original data size.

    Returns:
    tuple: A tuple containing the subset of features and labels.
    """

    unique_classes = np.unique(y)
    X_subset = []
    y_subset = []
    for class_value in unique_classes:
        X_class = X[y == class_value]
        y_class = y[y == class_value]
        X_class_subset, _, y_class_subset, _ = train_test_split(
            X_class, y_class, test_size=1-subset_ratio, stratify=y_class, random_state=42)
        X_subset.append(X_class_subset)
        y_subset.append(y_class_subset)
    X_subset = np.concatenate(X_subset, axis=0)
    y_subset = np.concatenate(y_subset, axis=0)
    return X_subset, y_subset

# Function to train the model
def train_model(X_train, y_train, X_dev, y_dev, subset_description, label_encoder):
    """
    Train the CNN model with the given training and validation data.

    Parameters:
    X_train (numpy.ndarray): Training data features.
    y_train (numpy.ndarray): Training data labels.
    X_dev (numpy.ndarray): Validation data features.
    y_dev (numpy.ndarray): Validation data labels.
    subset_description (str): Description of the training data subset.
    label_encoder (sklearn.preprocessing.LabelEncoder): The label encoder used for transforming class labels.
    """
    
    model = get_model()
    history = model.fit(
        X_train, y_train, 
        epochs=50, 
        batch_size=256, 
        callbacks=[LearningRateScheduler(lr_schedule)],
        validation_data=(X_dev, y_dev)
    )
    model_path = f'../trained_classifiers/fruit_classifier_{subset_description}.keras'
    model.save(model_path)

    plt.figure(figsize=(12, 4))
    plt.subplot(1, 2, 1)
    plt.plot(history.history['loss'], label='Training Loss')
    plt.plot(history.history['val_loss'], label='Validation Loss')
    plt.title(f'Loss Over Epochs ({subset_description})')
    plt.xlabel('Epochs')
    plt.ylabel('Loss')
    plt.legend()
    plt.savefig(f'../figures/cnn/fruit_classifier_{subset_description}/loss_plot_{subset_description}.png')

    plt.subplot(1, 2, 2)
    plt.plot(history.history['accuracy'], label='Training Accuracy')
    plt.plot(history.history['val_accuracy'], label='Validation Accuracy')
    plt.title(f'Accuracy Over Epochs ({subset_description})')
    plt.xlabel('Epochs')
    plt.ylabel('Accuracy')
    plt.legend()
    plt.savefig(f'../figures/cnn/fruit_classifier_{subset_description}/accuracy_plot_{subset_description}.png')

# Main execution block
if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Train a CNN on image data.')
    parser.add_argument('--train_subsets', nargs='+', type=float, default=[0.1], help='List of proportions of training data to use')
    args = parser.parse_args()

    train_subsets = args.train_subsets

    # Load and preprocess data
    X_train, y_train = load_images_and_labels_cnn('../data/train')
    X_dev, y_dev = load_images_and_labels_cnn('../data/dev')

    label_encoder = LabelEncoder()
    y_train_encoded = to_categorical(label_encoder.fit_transform(y_train))
    y_dev_encoded = to_categorical(label_encoder.transform(y_dev))

    for train_subset in train_subsets:
        subset_description = f'{int(train_subset*100)}_percent'
        if train_subset < 1.0:
            # Create stratified subset of data
            X_train_subset, y_train_subset_integers = stratified_subset(X_train, label_encoder.transform(y_train), train_subset)
            y_train_subset_encoded = to_categorical(y_train_subset_integers)
        else:
            X_train_subset = X_train
            y_train_subset_encoded = y_train_encoded

        # Train model
        train_model(X_train_subset, y_train_subset_encoded, X_dev, y_dev_encoded, subset_description, label_encoder)

    joblib.dump(label_encoder, '../figures/cnn/label_encoder.pkl')
