# Imports
import numpy as np
import os
import cv2
import tensorflow as tf
import matplotlib.pyplot as plt
import seaborn as sns
from sklearn.metrics import accuracy_score, precision_score, recall_score, f1_score, confusion_matrix
from tensorflow.keras.utils import to_categorical
from sklearn.preprocessing import LabelEncoder

def create_directory(directory_name):
    """
    Create a directory if it doesn't already exist.

    Parameters:
    directory_name (str): The name or path of the directory to be created.
    """

    if not os.path.exists(directory_name):
        os.makedirs(directory_name)

def save_plot(figure, directory, file_name):
    """
    Save a matplotlib plot to a specified directory with a given file name.

    Parameters:
    figure (matplotlib.figure.Figure): The matplotlib figure object to be saved.
    directory (str): The directory where the figure will be saved.
    file_name (str): The name of the file in which the figure will be saved.
    """

    file_path = os.path.join(directory, file_name)
    figure.savefig(file_path)
    plt.close(figure)

def load_images_and_labels_cnn(base_path, size=(50, 50)):
    """
    Load images and their labels for CNN processing from a given base path.
    Images are resized to a specified size.

    Parameters:
    base_path (str): The path of the directory containing image folders.
    size (tuple): The desired size for resizing the images (width, height).

    Returns:
    tuple: A tuple containing arrays of images, labels, and file paths.
    """

    images = []
    labels = []
    file_paths = []  # Store file paths
    for folder in os.listdir(base_path):
        folder_path = os.path.join(base_path, folder)
        for image_file in os.listdir(folder_path):
            img_path = os.path.join(folder_path, image_file)
            img = cv2.imread(img_path)
            if img is not None:
                img = cv2.resize(img, size)
                images.append(img)
                labels.append(folder)
                file_paths.append(img_path)  # Append file path
            else:
                print("Error loading image: " + img_path)
    return np.array(images), np.array(labels), file_paths

def plot_metrics(y_true, y_pred, metric_name, title, directory, file_name, is_accuracy=False):
    """
    Plot and save a metric evaluation bar chart.

    Parameters:
    y_true (array): True labels.
    y_pred (array): Predicted labels by the model.
    metric_name (function): The metric function to evaluate (e.g., accuracy_score).
    title (str): Title of the plot.
    directory (str): Directory to save the plot.
    file_name (str): Name of the file to save the plot.
    is_accuracy (bool): Indicator whether the metric is accuracy or not. Default is False.
    """

    if is_accuracy:
        metric_value = metric_name(y_true, y_pred)
    else:
        metric_value = metric_name(y_true, y_pred, average='macro')
    
    figure = plt.figure()
    sns.barplot(x=[0], y=[metric_value])
    plt.title(f'{title}: {metric_value:.4f}')
    plt.ylabel(title)
    plt.ylim(0, 1)
    plt.xticks([0], [title])
    save_plot(figure, directory, f'{file_name}.png')

def display_misclassified_images(original_paths, X, y_true, y_pred, label_encoder, num_images=7, directory='', model_name=''):
    """
    Display and save a specified number of misclassified images.

    Parameters:
    original_paths (list): List of paths to the original images.
    X (array): The array of image data.
    y_true (array): True labels.
    y_pred (array): Predicted labels.
    label_encoder (LabelEncoder): Encoder used to transform labels.
    num_images (int): Number of misclassified images to display. Default is 7.
    directory (str): Directory to save the images.
    model_name (str): Name of the model used for prefixing saved images.
    """
    
    misclassified_indices = np.where(y_true != y_pred)[0]
    if len(misclassified_indices) < num_images:
        num_images = len(misclassified_indices)
    misclassified_samples = np.random.choice(misclassified_indices, num_images, replace=False)

    figure = plt.figure(figsize=(15, 5))
    for i, index in enumerate(misclassified_samples, 1):
        original_img = plt.imread(original_paths[index])
        plt.subplot(1, num_images, i)
        plt.imshow(original_img)
        plt.title(f'Pred: {label_encoder.inverse_transform([y_pred[index]])[0]}\nTrue: {label_encoder.inverse_transform([y_true[index]])[0]}')
        plt.axis('off')

    save_path = os.path.join(directory, f'misclassified_images_{model_name}.png')
    plt.savefig(save_path)
    plt.close(figure)


def evaluate_model_with_metrics(model_path, datasets, label_encoder):
    """
    Evaluate a model with various metrics and save the results.

    Parameters:
    model_path (str): Path to the trained model.
    datasets (dict): A dictionary containing datasets for evaluation.
    label_encoder (LabelEncoder): Encoder used to transform labels.
    """

    model = tf.keras.models.load_model(model_path)
    model_name = os.path.basename(model_path).split('.')[0]
    results_directory = f'../figures/cnn/{model_name}'
    create_directory(results_directory)

    for split_name, (X, y, paths) in datasets.items():
        y_true = np.argmax(y, axis=1)
        y_pred = np.argmax(model.predict(X), axis=1)

        # Metrics
        accuracy = accuracy_score(y_true, y_pred)
        precision = precision_score(y_true, y_pred, average='macro')
        recall = recall_score(y_true, y_pred, average='macro')
        f1 = f1_score(y_true, y_pred, average='macro')

        print(f"Evaluating on {split_name} data:")
        print(f"Accuracy: {accuracy:.4f}")
        print(f"Precision (Macro): {precision:.4f}")
        print(f"Recall (Macro): {recall:.4f}")
        print(f"F1 Score (Macro): {f1:.4f}")

        # Save metrics plots
        plot_metrics(y_true, y_pred, accuracy_score, "Accuracy", results_directory, f"accuracy_{split_name}", is_accuracy=True)
        plot_metrics(y_true, y_pred, precision_score, "Precision", results_directory, f"precision_{split_name}")
        plot_metrics(y_true, y_pred, recall_score, "Recall", results_directory, f"recall_{split_name}")
        plot_metrics(y_true, y_pred, f1_score, "F1 Score", results_directory, f"f1_{split_name}")


        # Confusion Matrix
        confusion_mtx = confusion_matrix(y_true, y_pred)
        figure = plt.figure(figsize=(10, 8))
        sns.heatmap(confusion_mtx, annot=True, fmt='d', cmap='Blues', xticklabels=label_encoder.classes_, yticklabels=label_encoder.classes_)
        plt.xlabel('Predicted')
        plt.ylabel('True')
        plt.title(f'Confusion Matrix - {split_name}')
        save_plot(figure, results_directory, f'confusion_matrix_{split_name}.png')

        # Display misclassified images
        if split_name == 'Test':
            print(f"Displaying and saving misclassified images from {split_name} data:")
            y_true = np.argmax(y, axis=1)
            y_pred = np.argmax(model.predict(X), axis=1)
            display_misclassified_images(paths, X, y_true, y_pred, label_encoder, directory=results_directory, model_name=split_name)


def main(subset):
    """
    Main function to evaluate a CNN model on a specific dataset subset.

    Parameters:
    subset (str): Subset of the training data used for the model (e.g., "10_percent").
    """
    
    # Construct the model file path
    model_path = f'../trained_classifiers/fruit_classifier_{subset}.keras'

    # Load and preprocess data (as in the training script)
    X_train, y_train, train_paths = load_images_and_labels_cnn('../data/train')
    X_dev, y_dev, dev_paths = load_images_and_labels_cnn('../data/dev')
    X_test, y_test, test_paths = load_images_and_labels_cnn('../data/test')

    label_encoder = LabelEncoder()
    y_train_encoded = to_categorical(label_encoder.fit_transform(y_train))
    y_dev_encoded = to_categorical(label_encoder.transform(y_dev))
    y_test_encoded = to_categorical(label_encoder.transform(y_test))

    # Dictionary of datasets
    datasets = {
        'Train': (X_train, y_train_encoded, train_paths),
        'Dev': (X_dev, y_dev_encoded, dev_paths),
        'Test': (X_test, y_test_encoded, test_paths)
    }

    evaluate_model_with_metrics(model_path, datasets, label_encoder)

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Evaluate a CNN model on a specific dataset subset.')
    parser.add_argument('subset', type=str, help='Subset of the training data used for the model (e.g., "10_percent").')
    args = parser.parse_args()

    main(args.subset)

